@@ -5,11 +5,7 @@
  *
  */
 
-#include <linux/config.h>
-#ifdef MODULE
-#include <linux/module.h>
-#include <linux/version.h>
-#endif
+/* #include <linux/module.h>*/
 
 #include <linux/sched.h>
 #include <linux/errno.h>
@@ -22,8 +18,11 @@
 #include <linux/errno.h>
 #include "ncplib.h"
 
-#define NAME_OFFSET(de) ((int) ((de)->d_name - (char *) (de)))
-#define ROUND_UP(x) (((x)+3) & ~3)
+struct ncp_dirent {
+	struct nw_info_struct i;
+	struct nw_search_sequence s; /* given back for i */
+	unsigned long f_pos;
+};
 
 static int 
 ncp_dir_read(struct inode *inode, struct file *filp, char *buf, int count);
@@ -39,22 +38,11 @@ static int
 ncp_do_readdir(struct ncp_server *server, struct inode *dir, int fpos,
 	       int cache_size, struct ncp_dirent *entry);
 
-static int 
-get_pname(struct inode *dir, const char *name, int len,
-          char **res_path, int *res_len);
-
-static int
-get_pname_static(struct inode *dir, const char *name, int len,
-                 char *path, int *res_len);
-
 static struct inode *
-ncp_iget(struct inode *dir, char *path, struct ncp_dirent *finfo);
-
-static void 
-put_pname(char *path);
+ncp_iget(struct inode *dir, struct nw_file_info *finfo);
 
 static struct ncp_inode_info *
-ncp_find_inode(struct ncp_server *server, const char *path);
+ncp_find_inode(struct inode *dir, const char *name);
 
 static int
 ncp_lookup(struct inode *dir, const char *__name,
@@ -77,14 +65,6 @@ static int
 ncp_rename(struct inode *old_dir, const char *old_name, int old_len, 
            struct inode *new_dir, const char *new_name, int new_len);
 
-static int
-date_dos2unix(unsigned short time,unsigned short date);
-
-/*
-static void
-date_unix2dos(int unix_date,unsigned short *time, unsigned short *date);
-*/
-
 static inline void
 str_upper(char *name)
 {
@@ -162,6 +142,7 @@ ncp_readdir(struct inode *inode, struct file *filp,
         int index = 0;
 	struct ncp_dirent *entry = NULL;
         struct ncp_server *server = NCP_SERVER(inode);
+	struct ncp_inode_info *dir = (struct ncp_inode_info *)(inode->i_ino);
 
 	DDPRINTK("ncp_readdir: filp->f_pos = %d\n", (int)filp->f_pos);
 	DDPRINTK("ncp_readdir: inode->i_ino = %ld, c_ino = %ld\n",
@@ -180,25 +161,25 @@ ncp_readdir(struct inode *inode, struct file *filp,
 			printk("ncp_readdir: no MEMORY for cache\n");
 			return -ENOMEM;
 		}
-		for (i = 0; i < NCP_READDIR_CACHE_SIZE; i++) {
-			c_entry[i].path =
-                                (char *) ncp_kmalloc(NCP_MAXNAMELEN + 1,
-                                                     GFP_KERNEL);
-                        if (c_entry[i].path == NULL) {
-                                DPRINTK("ncp_readdir: could not alloc path\n");
-				while (--i>=0)
-					kfree(c_entry[i].path);
-				kfree(c_entry);
-				c_entry = NULL;
-				return -ENOMEM;
-                        }
-                }
 	}
 
         if (filp->f_pos == 0) {
                 ncp_invalid_dir_cache(inode->i_ino);
+		if (filldir(dirent,".",1, filp->f_pos, (int)dir) < 0) {
+			return 0;
+		}
+		filp->f_pos += 1;
         }
 
+	if (filp->f_pos == 1) {
+		if (filldir(dirent,"..",2, filp->f_pos, (int)(dir->dir)) < 0) {
+			return 0;
+		}
+		filp->f_pos += 1;
+	}
+
+	
+
 	if (inode->i_ino == c_ino) {
 		for (i = 0; i < c_size; i++) {
 			if (filp->f_pos == c_entry[i].f_pos) {
@@ -244,7 +225,7 @@ ncp_readdir(struct inode *inode, struct file *filp,
                         index = 0;
 
 			for (i = 0; i < c_size; i++) {
-				str_lower(c_entry[i].path);
+				str_lower(c_entry[i].i.entryName);
 			}
 		}
 	}
@@ -261,18 +242,9 @@ ncp_readdir(struct inode *inode, struct file *filp,
                    use. Otherwise the inode number does not
                    matter. (You can argue a lot about this..) */ 
 
-                int path_len;
-                int len;
                 struct ncp_inode_info *ino_info;
-                char complete_path[NCP_MAXPATHLEN];
-
-		len = strlen(entry->path);
-                if ((result = get_pname_static(inode, entry->path, len,
-                                               complete_path,
-                                               &path_len)) < 0)
-                        return result;
-
-                ino_info = ncp_find_inode(server, complete_path);
+                ino_info = ncp_find_inode(inode, entry->i.entryName);
+					  
 
                 /* Some programs seem to be confused about a zero
                    inode number, so we set it to one.  Thanks to
@@ -281,10 +253,10 @@ ncp_readdir(struct inode *inode, struct file *filp,
                         ino_info = (struct ncp_inode_info *) 1;
                 }
 
-		DDPRINTK("ncp_readdir: entry->path = %s\n", entry->path);
+		DDPRINTK("ncp_readdir: entry->path= %s\n", entry->i.entryName);
 		DDPRINTK("ncp_readdir: entry->f_pos = %ld\n", entry->f_pos);
 
-                if (filldir(dirent, entry->path, len,
+                if (filldir(dirent, entry->i.entryName, entry->i.nameLen,
                             entry->f_pos, (ino_t)ino_info) < 0) {
                         break;
                 }
@@ -307,29 +279,15 @@ ncp_read_volume_list(struct ncp_server *server, int fpos, int cache_size)
 {
 	struct ncp_dirent *entry = c_entry;
 
-	int total_count = 0;
+	int total_count = 2;
 	int i;
 
-	void fill_one(char *name)
-	{
-		if (total_count < fpos) {
-			DPRINTK("ncp_do_readdir: skipped file: %s\n", name);
-		} else if (total_count >= fpos + cache_size) {
-			return;
-		} else {
-			DPRINTK("ncp_do_readdir: found file: %s\n", name);
-			
-			entry->attr  = aDIR;
-			entry->mtime = 0;
-			entry->ctime = 0;
-			entry->atime = 0;
-			entry->size  = 1024;
-			entry->f_pos = total_count;
-			strcpy(entry->path, name);
-			entry += 1;
-		}
-		total_count += 1;
+#if 1
+	if (fpos < 2) {
+		printk("OOPS, we expect fpos >= 2");
 	}
+	fpos = 2;
+#endif
 
 	for (i=0; i<NCP_NUMBER_OF_VOLUMES; i++) {
 
@@ -340,11 +298,30 @@ ncp_read_volume_list(struct ncp_server *server, int fpos, int cache_size)
 		}
 
 		if (strlen(info.volume_name) > 0) {
-			fill_one(info.volume_name);
+			if (total_count < fpos) {
+				DPRINTK("ncp_read_volumes: skipped vol: %s\n",
+					info.volume_name);
+			} else if (total_count >= fpos + cache_size) {
+				return (total_count - fpos);
+			} else {
+				DPRINTK("ncp_read_volumes: found vol: %s\n",
+					info.volume_name);
+
+				if (ncp_do_lookup(server, NULL,
+						  info.volume_name,
+						  &(entry->i)) != 0) {
+					printk("ncpfs: could not lookup vol "
+					       "%s\n", info.volume_name);
+					continue;
+				}
+
+				entry->f_pos = total_count;
+				entry += 1;
+			}
+			total_count += 1;
 		}
 	}
-	fill_one(".");
-	fill_one("..");
+
 	return (total_count - fpos);
 }
 
@@ -352,73 +329,41 @@ static int
 ncp_do_readdir(struct ncp_server *server, struct inode *dir, int fpos,
 	       int cache_size, struct ncp_dirent *entry)
 {
-	struct ncp_filesearch_info fsinfo;
-	struct ncp_file_info       finfo;
+	struct nw_search_sequence seq;
+	int total_count = 2;
 
-	int total_count = 0;
+#if 1
+	if (fpos < 2) {
+		printk("OOPS, we expect fpos >= 2");
+	}
+	fpos = 2;
+#endif
+
+	if (ncp_initialize_search(server, NCP_ISTRUCT(dir), &seq) != 0) {
+		DPRINTK("ncp_init_search failed\n");
+		return total_count - fpos;
+	}
+
+	while (total_count < fpos + cache_size) {
+
+		if (ncp_search_for_file_or_subdir(server, &seq,
+						  &(entry->i)) != 0) {
+			return total_count - fpos;
+		}
 
-	void fill_one(int attr, char *name,
-		      int upd_time, int upd_date,
-		      int cr_date, int ac_date, int length)
-	{
 		if (total_count < fpos) {
-			DPRINTK("ncp_do_readdir: skipped file: %s\n", name);
-		} else if (total_count >= fpos + cache_size) {
-			return;
+			DPRINTK("ncp_do_readdir: skipped file: %s\n",
+				entry->i.entryName);
 		} else {
-			DPRINTK("ncp_do_readdir: found file: %s\n", name);
-			
-			entry->attr  = attr;
-			if ((attr & aDIR) == 0) {
-				entry->mtime = date_dos2unix(upd_time,
-							     upd_date);
-				entry->ctime = date_dos2unix(0, cr_date);
-				entry->atime = date_dos2unix(0, ac_date);
-				entry->size  = length;
-			} else {
-				/* Sorry, I do not know how to get the
-				 * values for directories :-(. Mail
-				 * lendecke@namu01.gwdg.de if you
-				 * know more. */
-				entry->mtime = 0;
-				entry->ctime = 0;
-				entry->atime = 0;
-				entry->size = 1024;
-			}
+			DPRINTK("ncp_do_readdir: found file: %s\n",
+				entry->i.entryName);
+			entry->s = seq;
 			entry->f_pos = total_count;
-			strcpy(entry->path, name);
 			entry += 1;
 		}
 		total_count += 1;
 	}
 
-	void doit(int attr) {
-		
-		if (ncp_file_search_init(server, 0, NCP_FINFO(dir)->path,
-					 &fsinfo) != 0) {
-			DPRINTK("could not fs init\n");
-			return;
-		}
-		
-		while (ncp_file_search_continue(server, &fsinfo, attr, "*",
-						&finfo) == 0) {
-
-			fill_one(attr, finfo.file_name,
-				 finfo.update_time, finfo.update_date,
-				 finfo.creation_date, finfo.access_date,
-				 finfo.file_length);
-			
-
-		}
-		return;
-	}
-
-	fill_one(aDIR, ".", 0, 0, 0, 0, 0);
-	fill_one(aDIR, "..", 0, 0, 0, 0, 0);
-
-	doit(0);
-	doit(aDIR);
-
 	return (total_count - fpos);
 }
 
@@ -441,17 +386,11 @@ ncp_invalid_dir_cache(unsigned long ino)
 void
 ncp_free_dir_cache(void)
 {
-        int i;
-
         DPRINTK("ncp_free_dir_cache: enter\n");
         
         if (c_entry == NULL)
                 return;
 
-        for (i = 0; i < NCP_READDIR_CACHE_SIZE; i++) {
-                ncp_kfree_s(c_entry[i].path, NAME_MAX + 1);
-        }
-
         ncp_kfree_s(c_entry,
                     sizeof(struct ncp_dirent) * NCP_READDIR_CACHE_SIZE);
         c_entry = NULL;
@@ -460,127 +399,8 @@ ncp_free_dir_cache(void)
 }
 
 
-/* get_pname_static: it expects the res_path to be a preallocated
-   string of len NCP_MAXPATHLEN. */
-static int
-get_pname_static(struct inode *dir, const char *name, int len,
-                 char *path, int *res_len)
-{
-        char *parentname = NCP_INOP(dir)->finfo.path;
-        int   parentlen  = NCP_INOP(dir)->finfo.len;
-
-#if 1
-        if (parentlen != strlen(parentname)) {
-                printk("get_pname: parent->finfo.len = %d instead of %d\n",
-                       parentlen, strlen(parentname));
-                parentlen = strlen(parentname);
-        }
-	
-#endif
-	DDPRINTK("get_pname_static: parentname = %s, parlen = %d\n",
-                 parentname, parentlen);
-	DDPRINTK("get_pname_static: name = %s, len = %d\n",
-		 name, len);
-	
-        if (len > NCP_MAXNAMELEN) {
-                return -ENAMETOOLONG;
-        }
-
-	/* Fast cheat for . */
-	if (len == 0 || (len == 1 && name[0] == '.')) {
-
-		memcpy(path, parentname, parentlen + 1);
-		*res_len = parentlen;
-		return 0;
-	}
-	
-	/* Hmm, what about .. ? */
-	if (len == 2 && name[0] == '.' && name[1] == '.') {
-
-		char *pos = strrchr(parentname, '\\');
-
-                if (   (pos == NULL)
-                    && (   (parentname[parentlen-1] == ':')
-			|| (parentlen == 0))) {
-
-                        /* We're at the top */
-
-                        path[0] = '\0';
-                        *res_len = 0;
-                        return 0;
-                }
-
-                
-		if (pos == NULL) {
-			printk("ncp_make_name: Bad parent NCP-name: %s",
-                               parentname);
-			return -ENODATA;
-		}
-		
-		len = pos - parentname;
-
-	        memcpy(path, parentname, len);
-		path[len] = '\0';
-	}
-	else if (parentlen == 0) {
-		memcpy(path, name, len);
-		path[len]   = ':';
-		path[len+1] = '\0';
-		len = len+1;
-	}
-	else
-	{
-		if (len + parentlen + 2 > NCP_MAXPATHLEN) 
-			return -ENAMETOOLONG;
-				
-		memcpy(path, parentname, parentlen);
-		path[parentlen] = '\\';
-		memcpy(path + parentlen + 1, name, len);
-		path[parentlen + 1 + len] = '\0';
-		len = parentlen + len + 1;
-	}
-
-	*res_len = len;
-
-	DDPRINTK("get_pname: path = %s, *pathlen = %d\n",
-                 path, *res_len);
-	return 0;
-}
-        
-static int 
-get_pname(struct inode *dir, const char *name, int len,
-          char **res_path, int *res_len)
-{
-        char result[NCP_MAXPATHLEN];
-        int  result_len;
-        int  res;
-
-        if ((res = get_pname_static(dir,name,len,result,&result_len)) != 0) {
-                return res;
-        }
-
-        if ((*res_path = ncp_kmalloc(result_len+1, GFP_KERNEL)) == NULL) {
-                printk("get_pname: Out of memory while allocating name.");
-                return -ENOMEM;
-        }
-
-        strcpy(*res_path, result);
-        *res_len = result_len;
-        return 0;
-}
-
-static void
-put_pname(char *path)
-{
-	ncp_kfree_s(path, 0);
-}
-
-/* Insert a NEW ncp_inode_info into the inode tree of our filesystem,
-   under dir. The caller must assure that it's not already there. We
-   assume that path is allocated for us. */
-
 static struct inode *
-ncp_iget(struct inode *dir, char *path, struct ncp_dirent *finfo)
+ncp_iget(struct inode *dir, struct nw_file_info *finfo)
 {
 	struct inode *inode;
         struct ncp_inode_info *new_inode_info;
@@ -591,11 +411,6 @@ ncp_iget(struct inode *dir, char *path, struct ncp_dirent *finfo)
 		return NULL;
 	}
 
-        if (!path) {
-                printk("ncp_iget: path is NULL\n");
-                return NULL;
-        }
-
 	if (!finfo) {
 		printk("ncp_iget: finfo is NULL\n");
 		return NULL;
@@ -605,18 +420,15 @@ ncp_iget(struct inode *dir, char *path, struct ncp_dirent *finfo)
                                      GFP_KERNEL);
 
         if (new_inode_info == NULL) {
-                printk("ncp_iget: could not alloc mem for %s\n", path);
+                printk("ncp_iget: could not alloc mem for %s\n",
+		       finfo->i.entryName);
                 return NULL;
         }
 
         new_inode_info->state = INODE_LOOKED_UP;
         new_inode_info->nused = 0;
         new_inode_info->dir   = NCP_INOP(dir);
-
-        new_inode_info->finfo        = *finfo;
-        new_inode_info->finfo.opened = 0;
-        new_inode_info->finfo.path   = path;
-        new_inode_info->finfo.len    = strlen(path);
+        new_inode_info->finfo = *finfo;
 
         NCP_INOP(dir)->nused += 1;
 
@@ -654,10 +466,12 @@ ncp_free_inode_info(struct ncp_inode_info *i)
                 i->next->prev = i->prev;
                 i->prev->next = i->next;
 
-                ncp_kfree_s(i->finfo.path, i->finfo.len+1);
+		DDPRINTK("ncp_free_inode_info: freeing %s\n",
+			 i->finfo.i.entryName);
+
                 ncp_kfree_s(i, sizeof(struct ncp_inode_info));
 
-                if (dir == NULL) return;
+                if (dir == i) return;
 
                 (dir->nused)--;
                 i = dir;
@@ -668,23 +482,24 @@ void
 ncp_init_root(struct ncp_server *server)
 {
         struct ncp_inode_info *root = &(server->root);
+	struct nw_info_struct *i = &(root->finfo.i);
+	unsigned short dummy;
 
 	DPRINTK("ncp_init_root: server %s\n", server->m.server_name);
+	DPRINTK("ncp_init_root: i = %x\n", (int)i);
 
         root->finfo.opened = 0;
-	root->finfo.attr   = aDIR;
-	root->finfo.size   = 1024;
-	root->finfo.mtime  = 0;
-	root->finfo.ctime  = 0;
-	root->finfo.atime  = 0;
-
-	server->root_path  = '\0';
-	root->finfo.path   = &(server->root_path);
-	root->finfo.len    = 0;
+	i->attributes  = aDIR;
+	i->dataStreamSize = 1024;
+	ncp_date_unix2dos(0, &(i->creationTime), &(i->creationDate));
+	ncp_date_unix2dos(0, &(i->modifyTime), &(i->modifyDate));
+	ncp_date_unix2dos(0, &dummy, &(i->lastAccessDate));
+	i->nameLen = 0;
+        i->entryName[0] = '\0';
 
         root->state = INODE_LOOKED_UP;
         root->nused = 1;
-        root->dir   = NULL;
+        root->dir   = root;
         root->next = root->prev = root;
         return;
 }
@@ -716,17 +531,20 @@ ncp_free_all_inodes(struct ncp_server *server)
 /* We will search the inode that belongs to this name, currently by a
    complete linear search through the inodes belonging to this
    filesystem. This has to be fixed. */
-
 static struct ncp_inode_info *
-ncp_find_inode(struct ncp_server *server, const char *path)
+ncp_find_inode(struct inode *dir, const char *name)
 {
+	struct ncp_server *server = NCP_SERVER(dir);
+	struct nw_info_struct *dir_info = NCP_ISTRUCT(dir);
         struct ncp_inode_info *result = &(server->root);
 
-        if (path == NULL)
+        if (name == NULL) {
                 return NULL;
+	}
 
         do {
-                if (strcmp(result->finfo.path, path) == 0)
+		if (   (result->finfo.i.DosDirNum == dir_info->DosDirNum)
+		    && (strcmp(result->finfo.i.entryName, name) == 0))
                         return result;
                 result = result->next;
 
@@ -739,13 +557,10 @@ static int
 ncp_lookup(struct inode *dir, const char *__name, int len,
            struct inode **result)
 {
-	char *name = NULL;
-	struct ncp_dirent finfo;
+	struct nw_file_info finfo;
 	struct ncp_server *server;
         struct ncp_inode_info *result_info;
-	int error;
         int found_in_cache;
-	int path_len;
 
 	*result = NULL;
 
@@ -765,24 +580,23 @@ ncp_lookup(struct inode *dir, const char *__name, int len,
 		return 0;
 	}
 
-	/* Now we will have to build up an NCP filename. */
-	if ((error = get_pname(dir, __name, len, &name, &path_len)) < 0) {
+	/* ..and for .. */
+	if (len == 2 && __name[0] == '.' && __name[1] == '.') {
+		*result = iget(dir->i_sb, (int)(NCP_INOP(dir)->dir));
 		iput(dir);
-		return error;
+		if (*result == 0)
+			return -EACCES;
+		else
+			return 0;
 	}
 
-	DDPRINTK("ncp_lookup: get_pname for %s returned %d\n",
-		 __name, error);
-
-        result_info = ncp_find_inode(NCP_SERVER(dir), name);
+        result_info = ncp_find_inode(dir, __name);
 
         if (result_info != 0) {
 
                 if (result_info->state == INODE_CACHED)
                         result_info->state = INODE_LOOKED_UP;
 
-                put_pname(name);
-
                 /* Here we convert the inode_info address into an
                    inode number */
 
@@ -796,10 +610,6 @@ ncp_lookup(struct inode *dir, const char *__name, int len,
 		return 0;
         }
 
-	/* Ok, now we have made our name. We have to build a new
-           ncp_inode_info struct and insert it into the tree, if it is
-           a name that exists on the server */
-
         /* If the file is in the dir cache, we do not have to ask the
            server. */
 
@@ -812,17 +622,14 @@ ncp_lookup(struct inode *dir, const char *__name, int len,
                 i = first;
                 do {
                         DDPRINTK("ncp_lookup: trying index: %d, name: %s\n",
-                                i, c_entry[i].path);
-                        if (strcmp(c_entry[i].path, __name) == 0) {
+                                i, c_entry[i].i.entryName);
+                        if (strcmp(c_entry[i].i.entryName, __name) == 0) {
                                 DPRINTK("ncp_lookup: found in cache!\n");
-				finfo = c_entry[i];
-				finfo.path = NULL; /* It's not ours! */
+				finfo.i = c_entry[i].i;
 				found_in_cache = 1;
 				break;
                         }
                         i = (i + 1) % c_size;
-                        DDPRINTK("ncp_lookup: index %d, name %s failed\n",
-                                 i, c_entry[i].path);
                 } while (i != first);
         }
 
@@ -832,59 +639,29 @@ ncp_lookup(struct inode *dir, const char *__name, int len,
 
 		memcpy(this_name, __name, len);
 		this_name[len] = 0;
+		str_upper(this_name);
 
-		if (dir->i_ino == (int)&(server->root)) {
+		DDPRINTK("ncp_lookup: do_lookup on %s/%s\n",
+			 NCP_ISTRUCT(dir)->entryName, this_name);
 
-			/* We want to look up a volume. We only want
-                           to know whether it exists, nothing more. */
-			int vol_no;
-
-			DPRINTK("ncp_lookup: looking up volume %s\n",
-				this_name);
-
-			if (ncp_get_volume_number(server,this_name,
-						  &vol_no)!=0) {
-				put_pname(name);
-				iput(dir);
-				return -ENOENT;
-			}
-
-			finfo.attr    = aDIR;
-			finfo.mtime   = 0;
-			finfo.ctime   = 0;
-			finfo.atime   = 0;
-			finfo.size    = 1024;
-
-		} else {
-			struct ncp_file_info ninfo;
-			if (ncp_get_finfo(server, 0, NCP_FINFO(dir)->path,
-					  this_name, &ninfo) != 0) {
-				error = -ENOENT;
-			}
-
-			finfo.attr   = ninfo.file_attributes;
-			finfo.mtime  = date_dos2unix(ninfo.update_time,
-						     ninfo.update_date);
-			finfo.ctime  = date_dos2unix(0, ninfo.creation_date);
-			finfo.atime  = date_dos2unix(0, ninfo.access_date);
-			finfo.size   = (ninfo.file_attributes & aDIR) != 0
-				? 1024 : ninfo.file_length;
-		}
-
-                if (error < 0) {
-                        put_pname(name);
+		if (ncp_do_lookup(server,
+				  dir->i_ino == (int)&(NCP_SERVER(dir)->root)
+				  ? NULL : NCP_ISTRUCT(dir),
+				  this_name,
+				  &(finfo.i)) != 0) {
                         iput(dir);
-                        return error;
+                        return -ENOENT;
                 }
         }
 
-	if (!(*result = ncp_iget(dir, name, &finfo))) {
-		put_pname(name);
+	finfo.opened = 0;
+	str_lower(finfo.i.entryName);
+
+	if (!(*result = ncp_iget(dir, &finfo))) {
 		iput(dir);
 		return -EACCES;
 	}
 
-        DDPRINTK("ncp_lookup: %s => %lu\n", name, (unsigned long)result_info);
 	iput(dir);
 	return 0;
 }
@@ -893,11 +670,8 @@ static int
 ncp_create(struct inode *dir, const char *name, int len, int mode,
            struct inode **result)
 {
-	int error;
-	char *path = NULL;
-	struct ncp_dirent entry;
-	struct ncp_file_info finfo;
-	struct ncp_dirent *ino_finfo;
+	struct nw_file_info finfo;
+	__u8 _name[len+1];
 
 	*result = NULL;
 
@@ -907,39 +681,30 @@ ncp_create(struct inode *dir, const char *name, int len, int mode,
 		return -ENOENT;
 	}
 
-	/* Now we will have to build up an NCP filename. */
-	if ((error = get_pname(dir, name, len, &path, &len)) < 0) {
-		iput(dir);
-		return error;
-	}
+	strncpy(_name, name, len);
+	_name[len] = '\0';
+	str_upper(_name);
 
-        if (ncp_create_newfile(NCP_SERVER(dir), 0, path, 0, &finfo) != 0) {
-		put_pname(path);
+	if (ncp_open_create_file_or_subdir(NCP_SERVER(dir),
+					   NCP_ISTRUCT(dir), _name,
+					   OC_MODE_CREATE|OC_MODE_OPEN,
+					   0, AR_READ|AR_WRITE,
+					   &finfo) != 0) {
 		iput(dir);
 		return -EACCES;
 	}
 
-        entry.attr  = finfo.file_attributes;
-	entry.mtime = date_dos2unix(finfo.update_time,
-				    finfo.update_date);
-	entry.ctime = date_dos2unix(0, finfo.creation_date);
-	entry.atime = date_dos2unix(0, finfo.access_date);
-        entry.size  = 0;
-
         ncp_invalid_dir_cache(dir->i_ino);
 
-	if (!(*result = ncp_iget(dir, path, &entry)) < 0) {
-		ncp_close_file(NCP_SERVER(dir), finfo.file_id);
-		put_pname(path);
+	str_lower(finfo.i.entryName);
+	finfo.access = O_RDWR;
+
+	if (!(*result = ncp_iget(dir, &finfo)) < 0) {
+		ncp_close_file(NCP_SERVER(dir), finfo.file_handle);
 		iput(dir);
-		return error;
+		return -EINVAL;
 	}
 
-	ino_finfo = &(NCP_INOP(*result)->finfo);
-	ino_finfo->opened = 1;
-	ino_finfo->access = O_RDWR;
-	memcpy(&(ino_finfo->file_id), finfo.file_id, NCP_FILE_ID_LEN);
-	
 	iput(dir);
 	return 0;	
 }
@@ -948,7 +713,12 @@ static int
 ncp_mkdir(struct inode *dir, const char *name, int len, int mode)
 {
 	int error;
-	char path[NCP_MAXPATHLEN];
+	struct nw_file_info new_dir;
+	__u8 _name[len+1];
+
+	strncpy(_name, name, len);
+	_name[len] = '\0';
+	str_upper(_name);
 
 	if (!dir || !S_ISDIR(dir->i_mode)) {
 		printk("ncp_mkdir: inode is NULL or not a directory\n");
@@ -956,14 +726,13 @@ ncp_mkdir(struct inode *dir, const char *name, int len, int mode)
 		return -ENOENT;
 	}
 
-	/* Now we will have to build up an NCP filename. */
-	if ((error = get_pname_static(dir, name, len, path, &len)) < 0) {
-		iput(dir);
-		return error;
-	}
-
-	if ((error = ncp_create_directory(NCP_SERVER(dir), 0, path,
-					  0xff)) == 0) {
+	if (ncp_open_create_file_or_subdir(NCP_SERVER(dir),
+					   NCP_ISTRUCT(dir), _name,
+					   OC_MODE_CREATE, aDIR, 0xffff,
+					   &new_dir) != 0) {
+		error = -EACCES;
+	} else {
+		error = 0;
                 ncp_invalid_dir_cache(dir->i_ino);
         }
 
@@ -975,24 +744,32 @@ static int
 ncp_rmdir(struct inode *dir, const char *name, int len)
 {
 	int error;
-        char path[NCP_MAXPATHLEN];
+	__u8 _name[len+1];
 
 	if (!dir || !S_ISDIR(dir->i_mode)) {
 		printk("ncp_rmdir: inode is NULL or not a directory\n");
 		iput(dir);
 		return -ENOENT;
 	}
-	if ((error = get_pname_static(dir, name, len, path, &len)) < 0) {
-		iput(dir);
-		return error;
-	}
-        if (ncp_find_inode(NCP_SERVER(dir), path) != NULL) {
+        if (ncp_find_inode(dir, name) != NULL) {
                 error = -EBUSY;
-        } else {
-                if ((error = ncp_delete_directory(NCP_SERVER(dir),
-						  0, path)) == 0) {
+        }
+	else
+	{
+
+		strncpy(_name, name, len);
+		_name[len] = '\0';
+		str_upper(_name);
+
+                if ((error = ncp_del_file_or_subdir(NCP_SERVER(dir),
+						    NCP_ISTRUCT(dir),
+						    _name)) == 0) {
                         ncp_invalid_dir_cache(dir->i_ino);
 		}
+		else
+		{
+			error = -EINVAL;
+		}
         }
 	iput(dir);
 	return error;
@@ -1002,25 +779,33 @@ static int
 ncp_unlink(struct inode *dir, const char *name, int len)
 {
 	int error;
-	char path[NCP_MAXPATHLEN];
+	__u8 _name[len+1];
 
 	if (!dir || !S_ISDIR(dir->i_mode)) {
 		printk("ncp_unlink: inode is NULL or not a directory\n");
 		iput(dir);
 		return -ENOENT;
 	}
-	if ((error = get_pname_static(dir, name, len, path, &len)) < 0) {
-		iput(dir);
-		return error;
-	}
-        if (ncp_find_inode(NCP_SERVER(dir), path) != NULL) {
+        if (ncp_find_inode(dir, name) != NULL) {
                 error = -EBUSY;
-        } else {
-                if ((error = ncp_erase_file(NCP_SERVER(dir), 0, path,
-					    0)) == 0)
-                        ncp_invalid_dir_cache(dir->i_ino);
         }
+	else
+	{
 
+		strncpy(_name, name, len);
+		_name[len] = '\0';
+		str_upper(_name);
+
+                if ((error = ncp_del_file_or_subdir(NCP_SERVER(dir),
+						    NCP_ISTRUCT(dir),
+						    _name)) == 0) {
+                        ncp_invalid_dir_cache(dir->i_ino);
+		}
+		else
+		{
+			error = -EINVAL;
+		}
+        }
 	iput(dir);
 	return error;
 }
@@ -1030,7 +815,8 @@ ncp_rename(struct inode *old_dir, const char *old_name, int old_len,
            struct inode *new_dir, const char *new_name, int new_len)
 {
 	int res;
-	char old_path[NCP_MAXPATHLEN], new_path[NCP_MAXPATHLEN];
+	char _old_name[old_len+1];
+	char _new_name[new_len+1];
 
 	if (!old_dir || !S_ISDIR(old_dir->i_mode)) {
 		printk("ncp_rename: old inode is NULL or not a directory\n");
@@ -1044,28 +830,32 @@ ncp_rename(struct inode *old_dir, const char *old_name, int old_len,
                 goto finished;
 	}
 
-        res = get_pname_static(old_dir, old_name, old_len, old_path, &old_len);
-        if (res < 0) {
-                goto finished;
-	}
-
-        res = get_pname_static(new_dir, new_name, new_len, new_path, &new_len);
-	if (res < 0) {
-                goto finished;
-	}
-	
-        if (   (ncp_find_inode(NCP_SERVER(old_dir), old_path) != NULL)
-            || (ncp_find_inode(NCP_SERVER(new_dir), new_path) != NULL)) {
+        if (   (ncp_find_inode(old_dir, old_name) != NULL)
+            || (ncp_find_inode(new_dir, new_name) != NULL)) {
                 res = -EBUSY;
                 goto finished;
         }
 
-	res = ncp_rename_file(NCP_SERVER(old_dir), 0, old_path, 0,0, new_path);
+	strncpy(_old_name, old_name, old_len);
+	_old_name[old_len] = '\0';
+	str_upper(_old_name);
+
+	strncpy(_new_name, new_name, new_len);
+	_new_name[old_len] = '\0';
+	str_upper(_new_name);
+
+	res = ncp_ren_or_mov_file_or_subdir(NCP_SERVER(old_dir),
+					    NCP_ISTRUCT(old_dir), _old_name,
+					    NCP_ISTRUCT(new_dir), _new_name);
 
         if (res == 0) {
                 ncp_invalid_dir_cache(old_dir->i_ino);
                 ncp_invalid_dir_cache(new_dir->i_ino);
-        }		
+        }
+	else
+	{
+		res = -EACCES;
+	}
 	
  finished:
 	iput(old_dir); 
@@ -1083,13 +873,11 @@ static int day_n[] = { 0,31,59,90,120,151,181,212,243,273,304,334,0,0,0,0 };
 
 extern struct timezone sys_tz;
 
-/*
 static int
 utc2local(int time)
 {
         return time - sys_tz.tz_minuteswest*60;
 }
-*/
 
 static int
 local2utc(int time)
@@ -1099,8 +887,8 @@ local2utc(int time)
 
 /* Convert a MS-DOS time/date pair to a UNIX date (seconds since 1 1 70). */
 
-static int
-date_dos2unix(unsigned short time,unsigned short date)
+int
+ncp_date_dos2unix(unsigned short time,unsigned short date)
 {
 	int month,year,secs;
 
@@ -1115,9 +903,8 @@ date_dos2unix(unsigned short time,unsigned short date)
 
 
 /* Convert linear UNIX date to a MS-DOS time/date pair. */
-#if 0
-static void
-date_unix2dos(int unix_date,unsigned short *time, unsigned short *date)
+void
+ncp_date_unix2dos(int unix_date,unsigned short *time, unsigned short *date)
 {
 	int day,year,nl_day,month;
 
@@ -1139,4 +926,3 @@ date_unix2dos(int unix_date,unsigned short *time, unsigned short *date)
 	}
 	*date = nl_day-day_n[month-1]+1+(month << 5)+(year << 9);
 }
-#endif
